# RoBlog.pm -- Apache request handler for Roman's Blog
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
# more details.
# 
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307  USA.
#
package RoBlog::RoBlog;
use Socket;
use DBI;
use Encode;
use Apache2::Const;
use Apache2::Request;
use Apache2::Connection;
use Apache2::URI;
use HTML::Entities;
use RoBlog::Config;
use RoBlog::Tmpl;
use RoBlog::Pingback;
use RoBlog::Util;

use strict;
no warnings 'redefine';
use vars qw($VERSION $dbh $req);

$VERSION = "0.3";


sub handler {
	my $r = shift;
	$req = new Apache2::Request($r, DISABLE_UPLOADS => 0, POST_MAX => 1048576);
	$dbh = open_database($req);

	(my $t = $req->uri()) =~ s/^\Q$basepath\E\/*//;
	if ($t =~ m,^((index(\.html?)?)?)?$,)	{ return redirect("story/".newest_story_id()); }
	elsif ($t =~ m{^all(@(\d+))?$})			{ return show_list(undef, $2); }
	elsif ($t =~ m{^cat/(\d+)(@(\d+))?$})	{ return show_list($1, $3); }
	elsif ($t =~ m{^story/(\d+)$})			{ return show_story($1); }
	elsif ($t =~ m{^comm/(\d+)(,(\d+))?(@([[:xdigit:]]+))?$})
											{ return do_comment($1, $3, $5); }
	elsif ($t =~ m{^pic/(\d+)$})			{ return do_pic($1); }
	elsif ($t eq "search")					{ return do_search(); }
	elsif ($t eq "style.css")				{ return do_css(); }
	elsif ($t eq "pingback")				{ return recv_pingback($req,$dbh);}
	elsif ($t eq "feed")					{ return do_atomfeed(); }
	elsif ($t =~ m{^atom(@(\d+))?$})		{ return do_atom($2); }
	elsif (grep { $_ eq $t } keys %RoBlog::Tmpl::staticpage) {
		do_staticpage($t);
	}
	else {
		return DECLINED;
	}
}


sub header {
	my ($title) = @_;

	$req->content_type("text/html; charset=utf-8");
	$req->headers_out->set("X-Pingback" => "$baseurl/pingback");
	print subst($RoBlog::Tmpl::htmlhead, TITLE => $title);
	print $RoBlog::Tmpl::header;

	my $rows = dbget_mm(qq{ SELECT id,name,count(sid)
							FROM cat JOIN catmap ON id = cid
							GROUP BY id,name ORDER BY name });
	print subst($RoBlog::Tmpl::navigation,
				CATLINKS =>
				join("", map { subst($RoBlog::Tmpl::navcatlink,
									 LINK => "$basepath/cat/$_->{id}",
									 TEXT => "$_->{name} ($_->{count})")
							   } @$rows));
	print $RoBlog::Tmpl::mainstart;
}

sub footer {
	print $RoBlog::Tmpl::mainend, $RoBlog::Tmpl::footer;
}

sub redirect {
	$req->headers_out->set(Location => "$baseurl/$_[0]");
	return REDIRECT;
}

sub show_list {
	my ($cat, $start) = @_;
	my ($linkbase, $tit, %bymon);
	$start ||= 0;

	($tit, $linkbase) = !$cat ? ("Index", "all@") :
		(dbget_11(qq{SELECT name FROM cat WHERE id=?}, $cat)||"??", "cat/$cat@");
	my $rows = dbget_mm(qq{ SELECT id,date,heading
							FROM stories JOIN catmap ON id = sid }.
						($cat ? " WHERE cid = $cat " : "").
						qq{ ORDER BY date DESC });
	for(my ($i,$j) = ($start,0); $i<@$rows && $j<$items_per_page; ++$i,++$j) {
		my $mon = ($rows->[$i]->{date} =~ /^(\d{4}-\d{2})/)[0];
		push(@{$bymon{$mon}}, $rows->[$i]);
	}

	header("$blogtitle: $tit");
	mkpageindex($linkbase, $start, scalar(@$rows), $items_per_page);
	foreach (sort { $b cmp $a } keys %bymon) {
		print subst($RoBlog::Tmpl::storylist,
					MONTH => $_,
					ITEMS => join("", map {
						subst($RoBlog::Tmpl::storylistitem,
							  LINK => "$basepath/story/$_->{id}",
							  TIME => shortdate($_->{date}),
							  TEXT => $_->{heading});
					} @{$bymon{$_}}));
	}
	mkpageindex($linkbase, $start, scalar(@$rows), $items_per_page);
	footer();
	return OK;
}

sub newest_story_id {
	return dbget_11(qq{SELECT id FROM stories ORDER BY date DESC LIMIT 1});
}

sub show_story {
	my $id = shift;
	my ($pb_text, $mkc_text, $com_text, $i) = ("", "", "");

	my $ids = dbget_m1(qq{ SELECT id FROM stories ORDER BY date DESC });
	for($i = 0; $i < @$ids && $ids->[$i] != $id; ++$i) {}
	return DECLINED if $i >= @$ids;
	
	my $art = dbget_1m(qq{ SELECT date,modified,heading,text
						   FROM stories WHERE id = ? }, $id)
		or return DECLINED;

	header("$blogtitle: $art->{heading}");
	mkprevnext("story/", $ids, $i);
	
	my $rows = dbget_mm(qq{ SELECT date,link,title FROM pingbacks
							WHERE sid = ? ORDER BY date DESC }, $id);
	$pb_text = subst($RoBlog::Tmpl::pingbacks, PINGBACKS =>
					 join("", map { subst($RoBlog::Tmpl::pingback,
										  DATE => shortdate($_->{date}),
										  LINK => $_->{link},
										  TITLE => $_->{title} || $_->{link}) }
						  @$rows))
		if $rows && @$rows;
	$rows = dbget_m1(qq{ SELECT id FROM comments
						 WHERE sid = ? AND rid IS NULL
						 ORDER BY date DESC }, $id);
	$com_text = subst($RoBlog::Tmpl::comments, COMMENTS =>
					  join("", map { comments_rec($id, $_) } @$rows))
		if $rows && @$rows;
	print subst($RoBlog::Tmpl::story,
				HEADING => $art->{heading},
				DATE => shortdate($art->{date}),
				TEXT => text_postprocess($art->{text}),
				EDITED => shortdate($art->{date}) ne shortdate($art->{modified}) ?
						  subst($RoBlog::Tmpl::edited,
								DATE => shortdate($art->{modified})) : "",
				PINGBACKS => $pb_text,
				MAKECOMMENT => subst($RoBlog::Tmpl::makecomment,
									 LINK => "$basepath/comm/$id"),
				COMMENTS => $com_text);
	mkprevnext("story/", $ids, $i);
	footer();
	return OK;
}

sub comments_rec {
	my ($sid, $id) = @_;

	my $row = dbget_1m(qq{ SELECT * FROM comments WHERE id = ? }, $id);
	my $email = ($row->{email} ?
				 subst($RoBlog::Tmpl::email_in_paren,
					   ADDR => $row->{email}, NAME => $row->{email}) : "");

	my $subcomm = dbget_m1(qq{ SELECT id FROM comments
							   WHERE rid = ? ORDER BY date DESC }, $id);
	my $subcomments = join("", map { comments_rec($sid, $_) } @$subcomm);
	return subst($RoBlog::Tmpl::comment,
				 DATE => shortdate($row->{date}),
				 NAME => $row->{name},
				 EMAIL => $email,
				 TEXT => $row->{text},
				 MAKECOMMENT => subst($RoBlog::Tmpl::makecomment,
									  LINK => "$basepath/comm/$sid,$id"),
				 SUBCOMMENTS => $subcomments);
}

sub do_comment {
	my ($sid, $cid, $cred) = @_;
	my $url = "comm/$sid".($cid ? ",$cid" : "");
	my $ref = $req->headers_in->{Referer};

	return($ref ne "$baseurl/story/$sid" ? DECLINED :
		   redirect("$url\@".comment_cred("$baseurl/$url")))
		if !$cred;
	$ref =~ /^([^@]+)(@([:xdigit:]+))?$/;
	!$3 || ($1 eq "$baseurl/$url" && $cred eq comment_cred($ref))
		or return htmlerror("401 Bad credential");
	if ($req->param && $req->param->{name} && $req->param->{text}) {
		my ($name, $mail, $text) = map {
			encode_utf8(encode_entities($req->param->{$_}, '<>&"')); #"
		} qw(name email text);
		$text =~ s,\n,<br/>,g;
		$dbh->do(qq{ INSERT INTO comments (id,sid,rid,name,email,text)
					 VALUES (DEFAULT,?,?,?,?,?) }, {},
				 $sid, $cid, $name, $mail, $text);
		return redirect("story/$sid");
	}
	
	my $oldtext = $cid ?
		dbget_11(qq{ SELECT text FROM comments WHERE id = ? }, $cid) :
		dbget_11(qq{ SELECT text FROM stories WHERE id = ? }, $sid);
	
	header("$blogtitle: $RoBlog::Tmpl::mkcommtitle");
	print subst($RoBlog::Tmpl::mkcommpage,
				LINK => "$baseurl/$url\@".
						comment_cred("$protocol://$hostname".$req->uri()),
				NAME => $req->param ? $req->param->{name} : "",
				NAMEERROR => ($req->param && !$req->param->{name} ?
							  $RoBlog::Tmpl::mkcommerror : ""),
				EMAIL => $req->param ? $req->param->{email} : "",
				TEXT => $req->param ? $req->param->{text} : "",
				TEXTERROR => ($req->param && !$req->param->{text} ?
							  $RoBlog::Tmpl::mkcommerror : ""),
				ORIGTEXT => $oldtext);
	footer();
	return OK;
}

sub do_pic {
	my $id = shift;
	
	my $pic = dbget_1m(qq{ SELECT fmt,pic FROM images WHERE id=? }, $id)
		or return DECLINED;
	$req->content_type("image/$pic->{fmt}");
	print $pic->{pic};
	return OK;
}

sub do_css {
	$req->content_type("text/css; charset=utf-8");
	print $RoBlog::Tmpl::css_style;
	return OK;
}

sub do_staticpage {
	my $name = shift;

	my $title = "$blogtitle";
	$title .= ": $RoBlog::Tmpl::staticpagetitle{$name}"
		if $RoBlog::Tmpl::staticpagetitle{$name};
	header($title);
	print $RoBlog::Tmpl::staticpage{$name};
	footer();
	return OK;
}

sub do_search {
	header("$blogtitle: $RoBlog::Tmpl::searchtitle");
	my $results = "";
	if ($req->param && $req->param->{text}) {
		(my $s = $req->param->{text}) =~ s/([_%])/\\$1/g;
		my $rows = dbget_mm(qq{ SELECT id,date,heading,text FROM stories
								WHERE heading ILIKE ? OR text ILIKE ?
								LIMIT $max_search }, "\%$s%", "\%$s%");
		$results = !$rows || !@$rows ? subst($RoBlog::Tmpl::searchnoresults) :
			subst($RoBlog::Tmpl::searchresults,
				  RESULTS =>join("", map {
					  subst($RoBlog::Tmpl::searchresult,
							LINK => "$basepath/story/$_->{id}",
							DATE => shortdate($_->{date}),
							HEADING => $_->{heading},
							EXCERPT => excerpt($_->{text}, $req->param->{text}));
				  } @$rows));
	}
	print subst($RoBlog::Tmpl::searchpage,
				PREVPATTERN => $req->param ? $req->param->{text} : "",
				RESULTS => $results);
	footer();
	return OK;
}

sub do_atomfeed {
	my $items = dbget_mm(qq{ SELECT id,date,modified,heading,text
							 FROM stories ORDER BY date DESC
							 LIMIT $feed_items });
	my $lastmod = dbget_11(qq{ SELECT modified FROM stories
							   ORDER BY modified DESC LIMIT 1 });
	$req->content_type("text/xml; charset=utf-8");
	print subst($RoBlog::Tmpl::atomfeed,
				DATE => isodate($lastmod),
				TITLE => $blogtitle,
				FEED => atomfeedlink(),
				POST => atompostlink(),
				ENTRIES =>
				join("\n", map {
					subst($RoBlog::Tmpl::atomentry,
						  TITLE => $blogtitle,
						  DATE => isodate($_->{date}),
						  MDATE => isodate($_->{modified}),
						  HEADING => $_->{heading},
						  ID => $_->{id},
						  FEED => "",
						  POST => "",
						  EDIT => atomeditlink($_->{id}),
						  LINK => "$baseurl/story/$_->{id}",
						  # sorry, drivel can't grok <content src="..."> :(
						  CONTENT => #subst($RoBlog::Tmpl::atomcontentref,
									 #	   LINK => "$baseurl/story/$_->{id}"))}
									 subst($RoBlog::Tmpl::atomcontent,
										   TEXT => $_->{text}))}
					 @$items));
	return OK;
}

sub do_atom {
	my $id = shift;
	my $method = $req->method();

	my $is_noauth = 0;
	my $remip = inet_atoi($req->connection()->remote_ip());
	foreach (@$noauth_ips) {
		my ($ip, $len) = (m,^([[:xdigit:].:]+)/(\d+)$,);
		$is_noauth = 1, last if inet_match($remip, inet_atoi($ip), $len);
	}
	if (!$is_noauth) {
		my ($auth, $wsse);
		if (!(($auth = $req->headers_in->{Authorization}) &&
			  $auth =~ /^WSSE/ && $auth =~ /profile="UsernameToken"/i &&
			  ($wsse = $req->headers_in->{"X-WSSE"}) &&
			  $wsse =~ /^UsernameToken\s+(.*)$/ &&
			  wsse_check($1))) {
			$req->headers_out->set("WWW-Authenticate" =>
								   'WSSE realm="RoBlog", profile="UsernameToken"');
			return htmlerror("401 Unauthorized",
							 "Authorization is required for this URL.");
		}
	}
	
	return DECLINED
		if (!(grep { $method eq $_ } qw(GET PUT POST)) ||
			($method ne "POST" && !$id));
	if ($method eq "PUT" || $method eq "POST") {
		my ($data, $title, $text, $cat, $date, $oldtext, $catid, $pingbacks);
		{ local($/) = undef; $data = <STDIN>; }
		my $entry = xmlparse($data, "entry", [qw(content title)], []);
		$entry !~ /^ERROR ((.|\n)*)$/
			or return htmlerror("400 Bad atom format: parse error: $1");
		ref($entry->{content})
			or return htmlerror("400 Bad atom format: content missing");
		xmlparse("<content>$entry->{content}{_T}</content>", "content", [], "ANY")
			!~ /^ERROR (.*)$/m
			or return htmlerror("400 Content parse error: $1");
		($title = atomtext($entry->{title}{_A}{type}, $entry->{title}{_T}))
			!~ /^ERROR (.*)$/m
			or return htmlerror("400 Bad atom format: $1")
			if $entry->{title};
		$title = encode_entities($title, '<>&"'); #"
		($text = atomtext($entry->{content}{_A}{type}, $entry->{content}{_T}))
			!~ /^ERROR (.*)$/m
			or return htmlerror("400 Bad atom format: $1");
		$cat = ($entry->{category}{_A}{label} || $entry->{category}{_A}{term}||
				($text =~ /\[cat:*([^\]]+)\]/)[0] || undef);
		$date = ($entry->{created}{_T} || $entry->{issued}{_T} ||
				 ($text =~ /\[date:*([^\]]+)\]/)[0] || undef);
		if ($date) {
			$date =~ s/[+-]\d{2,4}$//;
			$date =~ /^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}(:\d{2})?$/
				or return htmlerror("400 Bad date format");
		}
		$method ne "POST" || $cat
			or return htmlerror("400 Bad atom format: category missing");
		$method ne "POST" || $title
			or return htmlerror("400 Bad atom format: title missing");

		$oldtext = dbget_11(qq{ SELECT text FROM stories
								WHERE id = ? }, $id)
			if $method eq "PUT";
		($text, $pingbacks) = text_preprocess($text, $oldtext);
		$text !~ /^ERROR (.*)$/m
			or return htmlerror("400 Bad atom format", $1);

		$dbh->begin_work();
		if ($cat &&
			!($catid = dbget_11(qq{SELECT id FROM cat WHERE name=?}, $cat))) {
				$dbh->do("INSERT INTO cat (name) VALUES (?)", {}, $cat);
				$catid = dbget_11(qq{SELECT id FROM cat WHERE name=?}, $cat);
		}

		if ($method eq "POST") {
			do {
				$id = int(rand(999999))+1;
			} while(dbget_11(qq{SELECT id FROM stories WHERE id=?},$id));
			if (!$date) {
				my @t = localtime;
				$date = sprintf("%04d-%02d-%02d %02d:%02d:%02d",
								$t[5]+1900,$t[4]+1,$t[3],$t[2],$t[1],$t[0]);
			}
			$dbh->do("INSERT INTO stories (id,date,heading,text) VALUES (?,?,?,?)",
					 {}, $id, $date, encode_utf8($title), encode_utf8($text));
			$dbh->do("INSERT INTO catmap (sid,cid) VALUES (?,?)",
					 {}, $id, $catid);
			vmsg("Received new post from ".$req->connection()->remote_ip().
				 ", story id is $id");
			$req->headers_out->set("Location" => "$baseurl/atom\@$id");
		}
		else {
			$dbh->do("UPDATE stories SET text=? WHERE id = ?",
					 {}, encode_utf8($text), $id);
			$dbh->do("UPDATE stories SET heading=? WHERE id = ?",
					 {}, encode_utf8($title), $id)
				if defined($title);
			$dbh->do("UPDATE stories SET date=? WHERE id = ?",
					 {}, $date, $id)
				if defined($date);
			$dbh->do("UPDATE catmap SET cid=? WHERE sid = ?", {}, $catid, $id)
				if $catid;
			vmsg("Received edit from $remip for story id is $id");
		}
		$dbh->commit();
		foreach (@$pingbacks) {
			send_pingback("$baseurl/story/$id", $_);
		}
		send_ping();
		# POST should return an <entry>, so fall through to GET part
		return OK if $method eq "PUT";
	}
	
	my $art = dbget_1m(qq{ SELECT date,modified,heading,text
						   FROM stories WHERE id = ? }, $id)
		or return DECLINED;
	$req->content_type("text/xml; charset=utf-8");
	print subst($RoBlog::Tmpl::atomentry,
				DATE => isodate($art->{date}),
				MDATE => isodate($art->{modified}),
				HEADING => $art->{heading},
				ID => $art->{id},
				FEED => atomfeedlink(),
				POST => atompostlink(),
				EDIT => atomeditlink($art->{id}),
				LINK => "$baseurl/story/$art->{id}",
				CONTENT => subst($RoBlog::Tmpl::atomcontent,
								 TEXT => $art->{text}));
	return OK;
}


sub mkpageindex {
	my ($base, $pos, $max, $step) = @_;
	return if $max <= $step;
	my $pr = $pos - $step; $pr = 0 if $pr < 0;
	my $ne = $pos + $step; $ne = $max-$step if $ne >= $max;
	my $nums = "";
	for(my $i = 0; $i < $max; $i += $step) {
		my $j = $i/$step+1;
		$nums .= (($i == $pos) ?
				  subst($RoBlog::Tmpl::pidxcurr, NUM => $j) :
				  subst($RoBlog::Tmpl::pidxnum, NUM => $j,
						LINK => "$basepath/$base$i"));
	}
	
	print subst($RoBlog::Tmpl::pageindex,
				PREVLINK => $pos > 0 ?
							subst($RoBlog::Tmpl::pidxprev,
								  LINK => "$basepath/$base$pr") : "",
				INDEX    => subst($RoBlog::Tmpl::pidxnums,
								  NUMS => $nums),
				NEXTLINK => $pos < $max-1 && $ne > 0 ? 
							subst($RoBlog::Tmpl::pidxnext,
								  LINK => "$basepath/$base$ne") : "");
}

sub mkprevnext {
	my ($base, $ids, $curr) = @_;

	print subst($RoBlog::Tmpl::pageindex,
				PREVLINK => $curr > 0 ?
							subst($RoBlog::Tmpl::pidxprev,
								  LINK => "$basepath/$base".($ids->[$curr-1])) : "",
				INDEX    => subst($RoBlog::Tmpl::pidxnums, NUMS => "&nbsp;"),
				NEXTLINK => $curr < @$ids-1 ?
							subst($RoBlog::Tmpl::pidxnext,
								  LINK => "$basepath/$base".($ids->[$curr+1])) : "");
}

1;
